##-*- Mode: CPerl -*-

## File: Pd/Patch/Item.pm
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description:
##  + Base class for Pd patch items
##======================================================================

package Pd::Patch::Item;
use IO::File;
#use Text::Wrap qw();

use strict;

##======================================================================
## Package Globals

## $WRAP_COLUMNS
##  + number of chars per line for saveLines()
##  + set this to undef or zero to disable wrapping
our $WRAP_COLUMNS = 70;

##--> TODO: figure out how to propagate @PROPERTIES & friends
##    along inheritance lines!
##    + allow package name aliasing via inheritance, e.g.
##      @Pd::Patch::canvas::ISA = qw(Pd::Patch::Item::N_canvas);

## \@PROPERTIES = $PACKAGE->PROPERTIES()
##  + generic property index->name resolution for configure() & friends
sub PROPERTIES { return [qw(decl type)]; }

## \%DEFAULTS = $PACKAGE->DEFAULTS()
##  + generic property defaults by name (name=>$default): MASTER
sub DEFAULTS { return {decl=>'#X', type=>'obj'}; }


## \@PROPERTIES = $PACKAGE->PROPERTIES()
##  + generic property index->name resolution for configure() & friends
our @PROPERTIES=qw();

## %PROPERTIES
##  + generic property name->index resolution for configure() & friends: SLAVE
our %PROPERTIES = qw();

## \%DEFAULTS = $PACKAGE->DEFAULTS()
##  + generic property defaults by name (name=>$default): MASTER
our %DEFAULTS=qw();

## @DEFAULTS
##  + generic property defaults by index: SLAVE
our @DEFAULTS = qw();

## undef =       $SUBCLASS->init()
## undef = Pd::Patch::Item::init($SUBCLASS)
##  + sets up property name<-->index resolution for $SUBCLASS
sub init {
  my $pkg = shift;
  my $properties = $pkg->PROPERTIES();
  my $defaults   = $pkg->DEFAULTS();
  my ($name,$i);
  foreach $i (0..$#$properties) {
    next if (!defined($name=$properties->[$i]));
    eval "\$${pkg}::PROPERTIES[$i]    = q($name);";
    eval "\$${pkg}::PROPERTIES{$name} = $i;";
    if ($defaults && defined($defaults->{$name})) {
      eval "\$${pkg}::DEFAULTS[$i]    = q($defaults->{$name});";
      eval "\$${pkg}::DEFAULTS{$name} = q($defaults->{$name});";
    }
  }
  eval "sub ${pkg}::PROPERTIES_byname   { return \\\%${pkg}::PROPERTIES; }";
  eval "sub ${pkg}::PROPERTIES_bynumber { return \\\@${pkg}::PROPERTIES; }";
  eval "sub ${pkg}::DEFAULTS_byname     { return \\\%${pkg}::DEFAULTS;   }";
  eval "sub ${pkg}::DEFAULTS_bynumber   { return \\\@${pkg}::DEFAULTS;   }";
}
__PACKAGE__->init();


##======================================================================
## Constructors, etc.

## $item = $CLASS_OR_OBJ->new(@item_pd_args);
##  + object structure: HASH-ref:
##    _pd => \item_@pd_args,     ##-- e.g. [qw(#X obj 33 44 osc~ 1;)]
##  + @item_pd_args may also contain zero or more property HASH-refs,
##    which are passed to $item->configure()
sub new {
  my $that = shift;
  ##
  ##-- get pd decl & type: check for pd subclasses
  my ($decl,$type) = (grep {!ref($_) || ref($_) ne 'HASH'} @_)[0,1];
  my $thatclass = ref($that)||$that;
  my ($class);
  if (defined($decl) && defined($type)) {
    my $suffix = $decl.'_'.$type;
    $suffix =~ s/\#//g;
    my @prefixes = ("Pd::Patch::", "Pd::Patch::Item::");
    my ($prefix);
    foreach $prefix (@prefixes) {
      $class = $prefix.$suffix;
      return $class->new(@_)
	if (UNIVERSAL::isa($class,$thatclass) && $class ne $thatclass);
    }
  }
  return $thatclass->_new(@_);
}

## $item = $CLASS_OR_OBJ->_new(@configure_hashes_or_pdargs);
sub _new {
  my $that = shift;
  ##
  ##-- split args
  my %configure = qw();
  my @pdargs    = qw();
  foreach (@_) {
    if (ref($_) && ref($_) eq 'HASH') { @configure{keys(%$_)} = values(%$_); }
    else { push(@pdargs,$_); }
  }
  my $item = bless {_pd=>[@pdargs]}, ref($that)||$that;
  $item->configure(%configure) if (%configure);
  return $item;
}

## $item = $item->clear();
##  + clears an item (makes it invalid)
sub clear {
  my $item = shift;
  my $pd = $item->{_pd} || [];
  @$pd   = qw();
  %$item = (_pd=>$pd);
  return $item;
}

## $item = $item->sanitize();
##  + sanitize item: instantiate defaults
##  + called by 'saveLines()'
sub sanitize {
  my $item = shift;
  my $defaults = $item->propertyDefaults();
  foreach (0..$#$defaults) {
    $item->{_pd}[$_] = $defaults->[$_] if (!defined($item->{_pd}[$_]));
  }
  return $item;
}

##======================================================================
## Class-level access

## $bool = $class_or_obj->visible()
##  + returns true if item is visible (e.g. should be allocated an integer id by canvas)
##  + default returns true
sub visible { return 1; }

## $item2 = $item->clone()
##  + clones the item (uses save/load lines)
##  + implicitly instantiates defaults!
sub clone {
  return $_[0]->loadLines($_[0]->saveLines);
}


##======================================================================
## Property-level access

## \@PROPERTIES = $CLASS_OR_OBJ->_PROPERTIES()
##  + returns properties array: inheritance-related

##  @i2name = $CLASS_OR_OBJ->propertyNames(); ##-- list context
## \@i2name = $CLASS_OR_OBJ->propertyNames(); ##-- scalar context
##  + get all known property names for an object, in .pd-file order
sub propertyNames {
  my $that = shift;
  return (wantarray
	  ?  @{ $that->PROPERTIES() }
	  :     $that->PROPERTIES()
	 );
}

##  %name2i = $CLASS_OR_OBJ->propertyIndices(); ##-- list context
## \%name2i = $CLASS_OR_OBJ->propertyIndices(); ##-- scalar context
##  + get all known property indices for an object, indexed by name
sub propertyIndices {
  my $that = shift;
  return (wantarray
	  ?  %{ $that->PROPERTIES_byname() }
	  :     $that->PROPERTIES_byname()
	 );
}

##  @i2default = $CLASS_OR_OBJ->propertyDefaults(); ##-- list context
## \@i2default = $CLASS_OR_OBJ->propertyDefaults(); ##-- scalar context
##  + get property defaults
sub propertyDefaults {
  my $that = shift;
  return (wantarray
	  ?  @{ $that->DEFAULTS_bynumber() }
	  :     $that->DEFAULTS_bynumber()
	 );
}

##  %name2default = $CLASS_OR_OBJ->propertyDefaultsHash(); ##-- list context
## \%name2default = $CLASS_OR_OBJ->propertyDefaultsHash(); ##-- scalar context
##  + get property defaults
sub propertyDefaultsHash {
  my $that = shift;
  return (wantarray
	  ?  %{ $that->DEFAULTS_byname() }
	  :     $that->DEFAULTS_byname()
	 );
}

##  @propertyIndices = $item->getPropertyIndices(@propertyNames); ##-- list context
## \@propertyIndices = $item->getPropertyIndices(@propertyNames); ##-- scalar context
##  + does some sanity checking
sub getPropertyIndices {
  my $that = shift;
  my $name2i = $that->propertyIndices();
  my @indices = qw();
  foreach (@_) {
    if (!defined($_)) {
      warn(ref($that), "::getPropertyIndices(): undefined property name");
      push(@indices,undef);
      next;
    }
    if (!defined($name2i->{$_})) {
      ##-- check for "property${i}" names
      if ($_ =~ /^property(\d+)$/) {
	push(@indices,$1);
      } else {
	warn(ref($that), "::getPropertyIndices(): unknown property name '$_'");
      }
      next;
    }
    push(@indices,$name2i->{$_});
  }
  return wantarray ? @indices : [@indices];
}


## @vals    = $item->cget($propertyName, ...);
## $vals[0] = $item->cget($propertyName);
sub cget {
  my $item = shift;
  my @vals = ( map { defined($_) ? $item->{_pd}[$_] : undef } $item->getPropertyIndices(@_) );
  return wantarray ? @vals : $vals[0];
}

##  %name2val = $item->configure($propertyName=>$propertyValue, ...); ##-- list context
## \%name2val = $item->configure($propertyName=>$propertyValue, ...); ##-- scalar context
*cset = \&configure;
sub configure {
  my ($item,%name2val) = @_;
  my @keys    = keys(%name2val);
  my @vals    = values(%name2val);
  my @indices = $item->getPropertyIndices(@keys) if (@keys);
  foreach (0..$#keys) {
    next if (!defined($indices[$_]));
    $item->{_pd}[$indices[$_]] = $vals[$_];
  }

  ##-- get full return configuration
  my %chash  = qw();
  my $i2name = $item->propertyNames();
  my ($name);
  foreach (0..($#{$item->{_pd}} > $#$i2name ? $#{$item->{_pd}} : $#$i2name)) {
    $name = defined($i2name->[$_]) ? $i2name->[$_] : "property$_";
    $chash{$name} = $item->{_pd}[$_] if (defined($item->{_pd}[$_]));
  }
  return wantarray ? %chash : \%chash;
}


##======================================================================
## I/O

##--------------------------------------------------------------
## I/O: Lines (array of strings)

## $itemString_or_undef = $CLASS_OR_OBJ->nextItemString(\@lines);
##  + low-level: reads next single item string from \@lines
##  + destructively alters \@lines
sub nextItemString {
  my ($that,$lines) = @_;
  return undef if (!$lines || !@$lines);
  my ($line);
  my $str = '';
  while (defined($line=shift(@$lines))) {
    $str .= $line;
    last if ($line =~ /[^\\]\;$/ || $line =~ /^\;$/);
  }
  $str =~ s/\;$//;
  return $str;
}

## $item = $item->loadLines(\@lines)
##  + load from an array of lines
##  + destructively alters \@lines
sub loadLines {
  my ($that,$lines) = @_;
  my $istr = $that->nextItemString($lines);
  return undef if (!defined($istr));

  my @pdargs = grep {defined($_) && $_ ne ''} split(/\s+/, $istr);
  my $item = $that->new(@pdargs);

  ##-- load more data
  return $item->loadData($lines) if ($item->can('loadData'));
  return $item;
}

## $item = $item->loadData(\@lines);
##  + load extra item data
sub loadData { return $_[0]; }

## \@lines = $item->saveLines(); ##-- scalar context
##   + save to an array of lines
sub saveLines {
  my $item = shift;
  $item->sanitize();
  return [$item->_wrap(join(' ', map {defined($_) ? $_ : 0} @{$item->{_pd}}) . ";\n")];
}

## $text_wrapped = $class_or_obj->_wrap($text);
## $text_wrapped = $class_or_obj->_wrap($text,$ncols);
##   + $cols defaults to $WRAP_COLUMNS
sub _wrap {
  my ($that,$text,$ncols) = @_;
  $ncols = $WRAP_COLUMNS if (!defined($ncols));
  return $text if (!$ncols || $ncols <= 0);
  my $wrapped = '';
  my $line = '';
  my ($token,$sep);
  while ($text =~ /(\s*)(\S+)/g) {
    ($sep,$token)=($1,$2);
    $line .= ($line ne '' ? $sep : '').$token;
    if (length($line) >= $ncols) {
      $wrapped .= $line."\n";
      $line     = '';
    }
  }
  $wrapped .= $line;
  return ($wrapped =~ /\n$/s ? $wrapped : $wrapped."\n");
}

sub _wrap0 {
  my ($that,$text,$ncols) = @_;
  $ncols = $WRAP_COLUMNS if (!defined($ncols));
  return $text if (!$ncols || $ncols <= 0);
  my $wrapped = '';
  my $line = '';
  my ($token,$sep);
  while ($text =~ /(\s*)(\S+)/g) {
    ($sep,$token)=($1,$2);
    if (!$line || length($line)+length($sep)+length($token) < $ncols) {
      $line .= $sep.$token;
    }
    else {
      $wrapped .= $line."\n";
      $line     = $token;
    }
  }
  $wrapped .= $line;
  return $wrapped."\n";
}


##--------------------------------------------------------------
## I/O: String

## $item = $CLASS_OR_OBJ->loadString($string)
##  + instantiates the *first* item from $item_string
##  + calls $item->loadLines(\@lines)
sub loadString {
  my ($that,$str) = @_;
  my @lines = split(/\n+/,$str);
  return $that->loadLines(\@lines);
}

## $str = $item->saveString()
##   + calls saveLines()
sub saveString {
  my $item = shift;
  return join('', @{$item->saveLines});
}

##--------------------------------------------------------------
## I/O: File

## $item = $CLASS_OR_OBJ->loadFile($filename_or_fh)
##   + loads the 1st item from $filename_or_fh
##   + calls $item->loadFh($fh)
sub loadFile {
  my ($that,$file) = @_;
  my $fh = ref($file) ? $file : IO::File->new("<$file");
  die(__PACKAGE__,"::loadFile(): open failed for '$file'") if (!$fh);

  my $item = $that->loadFh($fh);
  $fh->close() if (!ref($file));
  return $item;
}

## $bool = $CLASS_OR_OBJ->saveFile($filename_or_fh)
##   + saves item to $filename_or_fh (clobbers existing files)
##   + calls $item->saveFh($fh)
sub saveFile {
  my ($item,$file) = @_;
  my $fh = ref($file) ? $file : IO::File->new(">$file");
  die(__PACKAGE__,"::saveFile(): open failed for '$file'") if (!$fh);

  my $rc = $item->saveFh($fh);
  $rc &&= $fh->close() if (!ref($file));
  return $rc;
}

##--------------------------------------------------------------
## I/O: Fh

## $item_or_undef = $CLASS_OR_OBJ->loadFh($fh)
##   + loads the 1st item from $fh
##   + calls $CLASS_OR_OBJ->loadLines([<$fh>]);
sub loadFh {
  my ($that,$fh) = @_;
  return $that->loadLines([<$fh>]);
}

## $bool = $item->saveFh($fh)
##   + saves item to $fh
##   + calls $item->saveLines();
sub saveFh {
  my ($item,$fh) = @_;
  $fh->print(@{$item->saveLines()});
  return 1;
}


1; ##-- be happy

__END__

##======================================================================
## Docs
=pod

=head1 NAME

Pd::Patch - Perl parser and wrapper classes for Pd abstractions

=head1 SYNOPSIS

 use Pd::Patch;

 #... stuff happens...

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

not yet written

=cut

##======================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

Pd by Miller Puckette and others.

=head1 AUTHOR

Bryan Jurish E<lt>moocow@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2007, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1),
pd(1).

=cut
