##-*- Mode: CPerl -*-

## File: Pd/Patch/canvas.pm
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description:
##  + class for Pd canvas items (patches, subpatches)
##======================================================================

package Pd::Patch::N_canvas;
use Pd::Patch::Item;
use IO::File;

use strict;

##======================================================================
## Package Globals

our @ISA = qw(Pd::Patch::Item);


sub PROPERTIES {
  return [qw(decl type x y width height arg1)];
  ##-- arg1=fontsize (toplevel), name (subpatch)
}

sub DEFAULTS {
  return {decl=>'#N', type=>'canvas', 'x'=>0, 'y'=>0, width=>450, height=>300, unk1=>10};
}

__PACKAGE__->init();


##-- aliases
@Pd::Patch::canvas::ISA   = (__PACKAGE__);
@Pd::Patch::subpatch::ISA = (__PACKAGE__);


##======================================================================
## Constructors, etc.

## $c = $CLASS_OR_OBJ->new(@canvas_pd_args);
##  + object structure: HASH-ref:
##    _pd        => \@pd_form,       ##-- e.g. [qw(#X canvas 0 0 350 300 ? ?)]
##    _items     => \%items,         ##-- pseudo-set of contained items
##    _item2n    => \%item2i,        ##-- item creation order (may contain gaps)
##    _in        => $n,              ##-- item number of next item
##    _connect   => \%connections,   ##-- pseudo-set of canvas-local connections
##    _cn        => $n,              ##-- connection number of next connection
##    _connect2n => \%connect2n,     ##-- connection creation order
##    _coords    => \@coords_item,   ##-- coords pseudo-item (gop properties)
##    _restore   => \@restore_item,  ##-- restoration pseudo-item (only for subpatch canvases)
##
##  + %connections:
##      ( "${fromObj}#${fromOutlet}->${toObj}#${toInlet}" => undef, ... )
sub new {
  my $that = shift;
  my $c = $that->SUPER::_new(@_);
  $c->{_items}     = {} if (!$c->{_items});
  $c->{_in}        = 0  if (!$c->{_in});
  $c->{_item2n}    = { map {($_=>$c->{_in}++)} keys(%{$c->{_items}}) } if (!$c->{_item2n});
  $c->{_connect}   = {} if (!$c->{_connect});
  $c->{_cn}        = 0  if (!$c->{_cn});
  $c->{_connect2n} = { map {($_=>$c->{_cn}++)} keys(%{$c->{_connect}}) } if (!$c->{_connect2n});
  #$c->{_coords}    = undef;
  #$c->{_restore}   = undef;
  return $c;
}

## $item = $item->clear();
sub clear {
  my $c = shift;
  $c->SUPER::clear();
  %{$c->{_items}} = qw();
  $c->{_in} = 0;
  %{$c->{_item2n}} = qw();
  %{$c->{_connect}} = qw();
  $c->{_cn} = 0;
  %{$c->{_connect2n}} = qw();
  #$c->{_coords} = undef;
  #$c->{_restore} = undef;
  return $c;
}


## $item = $item->sanitize();
##  + sanitize item: instantiate defaults
##  + called by 'saveLines()': may instantiate $item->{iargs} for saving
sub sanitize {
  my $c = shift;
  @{$c->{_pd}}[0,1] = ('#N','canvas'); ##-- force
  return $c->SUPER::sanitize();
}


##======================================================================
## Utilities

## $cmp = $c->_itemcmp($i1,$i2);
##  + for creation-order item sorting
sub _itemcmp {
  my $c = shift;
  my ($an,$bn);
  return (defined($an=$c->{_item2n}{$a}) && defined($bn=$c->{_item2n}{$b})
	  ? ($an <=> $bn)
	  : 0);
}

## $cmp = $c->_connectcmp($a,$b)
##  + for creation-order connection sorting
sub _connectcmp {
  my $c = shift;
  my ($an,$bn);
  return (defined($an=$c->{_connect2n}{$a}) && defined($bn=$c->{_connect2n}{$b})
	  ? ($an <=> $bn)
	  : 0);
}

##  @items = $c->items(); ##-- list context
## \@items = $c->items(); ##-- array context
##  + returns all items, in creation order
sub items {
  my $c = shift;
  my @items = sort { $c->_itemcmp($a,$b) } values(%{$c->{_items}});
  return wantarray ? @items : \@items;
}

##  @connections = $c->connections(); ##-- list context
## \@connections = $c->connections(); ##-- array context
##  + returns all connections (parsed), in creation order
sub connections {
  my $c = shift;
  my @connections = (
		     map  { scalar($c->splitConnection($_)) }
		     sort { $c->_connectcmp($a,$b) }
		     keys %{$c->{_connect}}
		    );
  return wantarray ? @connections : \@connections;
}

## ($fromObj,$fromOutlet, $toObj,$toInlet) = $c->splitConnection($connection); ##-- list context
## [$fromObj,$fromOutlet, $toObj,$toInlet] = $c->splitConnection($connection); ##-- scalar context
##  + parses a connection specification
sub splitConnection {
  my ($c,$connect) = @_;
  my (@parsed);
  if ($connect =~ /^([^\#]+)\#(\d+)\-\>([^\#]+)\#(\d+)$/) {
    @parsed    = ($1,$2,$3,$4);
    $parsed[0] = $c->{_items}{$parsed[0]} if (defined($c->{_items}{$parsed[0]}));
    $parsed[2] = $c->{_items}{$parsed[2]} if (defined($c->{_items}{$parsed[2]}));
  }
  else {
    warn(__PACKAGE__, "::splitConnection(): could not parse connection '$connect'");
  }
  return wantarray ? @parsed : [@parsed];
}

## $connect_str = $CLASS_OR_OBJ->joinConnection($connect)
##  + unparses a connection
sub joinConnection {
  my ($c,$connect) = @_;
  return $connect if (!UNIVERSAL::isa($connect,'ARRAY'));
  return $connect->[0].'#'.($connect->[1]||0).'->'.$connect->[2].'#'.($connect->[3]||0);
}

##--------------------------------------------------------------
## Manipulation: Items

## $c = $c->insert(@items);
##  + add a new canvas item
sub insert {
  my $c = shift;
  my ($item);
  foreach $item (@_) {
    $c->{_items}{$item}  = $item;
    if ($item->visible()) {
      $c->{_item2n}{$item} = int($c->{_in});
      $c->{_in}            = int($c->{_in}) + 1;
    } else {
      $c->{_item2n}{$item} = $c->{_in};
      $c->{_in}           += 0.0001;
    }
  }
  return $c;
}

## undef = $c->remove(@items);
##  + removes item(s) and any connections to/from them
sub remove {
  my $c = shift;
  my %items = map {($_=>undef)} @_;
  delete(@{$c->{_items}}{@_});
  ##
  ##-- scan and remove connections: slow and icky
  my ($connect,@csplit);
  foreach $connect (keys(%{$c->{_connect}})) {
    @csplit = $c->splitConnection($connect);
    delete($c->{_connect}{$connect})
      if (exists($items{$csplit[0]}) || exists($items{$csplit[2]}));
  }
  return $c;
}

##  @items = $c->find(%conditions); ##-- list context
## \@items = $c->find(%conditions); ##-- scalar context
##  + find matching items
##  + conditions:
##     $configureKey => $configureVal, ##-- literally match any known property (conjoined)
##     match         => \&sub,         ##-- user-defined matching
##     isa           => $perl_class,   ##-- test class membership (with isa())
sub find {
  my $c = shift;
  my ($item,@matched,$rc,$i,$ckey,$cval,$itemval);
  foreach $item (@{$c->items}) {
    $rc = 1;
    for ($i=0; $rc && $i <= $#_; $i+=2) {
      ($ckey,$cval) = @_[$i,$i+1];
      next if (!defined($cval));

      if    ($ckey eq 'match') { $rc &&= $cval->($item); next; }
      elsif ($ckey eq 'isa')   { $rc &&= $item->isa($cval); next; }

      $itemval   = $item->cget($ckey);
      $rc      &&= (defined($itemval) && $itemval eq $cval);
    }
    push(@matched,$item) if ($rc);
  }
  return wantarray ? @matched : [@matched];
}

##--------------------------------------------------------------
## Manipulation: Connections

## @connections = $c->connectionsFrom($item);
## @connections = $c->connectionsFrom($item,$outlet);
##  + get all connections from $item (1st form) or $item[$outlet] (2nd form)
sub connectionsFrom {
  my ($c,$item,$outlet) = @_;
  my @connections = grep {
    ($_->[0] eq $item && (!defined($outlet) || $_->[1]==$outlet))
  } @{$c->connections()};
  return wantarray ? @connections : [@connections];
}

## @connections = $c->connectionsTo($item);
## @connections = $c->connectionsTo($item,$inlet);
##  + get all connections to $item (1st form) or $item[$inlet] (2nd form)
sub connectionsTo {
  my ($c,$item,$inlet) = @_;
  my @connections = grep {
    ($_->[2] eq $item && (!defined($inlet) || $_->[3]==$inlet))
  } @{$c->connections()};
  return wantarray ? @connections : [@connections];
}

## $connection = $c->connect($fromObj,$fromOutlet, $toObj,$toInlet)
##  + add a new connection
sub connect {
  my ($c,$fromObj,$fromOutlet,$toObj,$toInlet) = @_;
  if (!defined($c->{_items}{$fromObj})) {
    warn(ref($c), "::connect(): unknown fromObject '$fromObj'");
    return undef;
  }
  if (!defined($c->{_items}{$toObj})) {
    warn(ref($c), "::connect(): unknown toObj '$toObj'");
    return undef;
  }
  my $connect = $fromObj.'#'.($fromOutlet||0).'->'.$toObj.'#'.($toInlet||0);
  if (!exists($c->{_connect}{$connect})) {
    ##-- new connection
    $c->{_connect}{$connect} = undef;
    $c->{_connect2n}{$connect} = $c->{_cn}++;
  }
  return $connect;
}

## undef = $c->disconnect($fromObj,$fromOutlet, $toObj,$toInlet)
##  + remove a connection
sub disconnect {
  my ($c,$fromObj,$fromOutlet,$toObj,$toInlet) = @_;
  my $connect = $fromObj.'#'.($fromOutlet||0).'->'.$toObj.'#'.($toInlet||0);
  delete($c->{_connect}{$connect});
  return $connect;
}



##======================================================================
## I/O

##--------------------------------------------------------------
## I/O: Lines

## $canvas = $canvas->loadData(\@lines)
##  + loads canvas data
sub loadData {
  my ($c,$lines) = @_;
  my ($item);
  my @n2item = qw();
  my ($fromObjN,$fromOutlet, $toObjN,$toInlet, $connect);

  while ($lines && @$lines) {
    last if (!defined($item = Pd::Patch::Item->loadLines($lines)));

    ##-- check for specials
    if ($item->{_pd}[0] eq '#X') {

      ##-- check for & parse connections
      if ($item->{_pd}[1] eq 'connect') {
	($fromObjN,$fromOutlet, $toObjN,$toInlet) = @{$item->{_pd}}[2..5];
	$c->connect($n2item[$fromObjN],$fromOutlet, $n2item[$toObjN],$toInlet);
	next;
      }
      ##-- check for gop coords
      elsif ($item->{_pd}[1] eq 'coords') {
	$c->{_coords} = $item;
	next;
      }
      ##-- check for subpatch restore
      elsif ($item->{_pd}[1] eq 'restore') {
	$c->{_restore} = $item;
	last;
      }
    }

    ##-- add an item
    $c->insert($item);
    $n2item[$c->{_item2n}{$item}] = $item if ($item->visible);
  }
  return $c;
}


## $str = $item->saveLines()
##   + overridable
sub saveLines {
  my $c = shift;
  $c->sanitize();
  my $items    = $c->items();
  my %i2n      = (map { ($items->[$_]=>$_) } (0..$#$items));
  my $connects = $c->connections();
  my $connect  = undef;
  return [
	  @{$c->SUPER::saveLines()},
	  (map { @{$_->saveLines()} } @$items),
	  (map {
	    $connect = ref($_) ? $_ : $c->splitConnection($_);
	    "#X connect $i2n{$connect->[0]} $connect->[1] $i2n{$connect->[2]} $connect->[3];\n"
	  } @$connects),
	  (defined($c->{_coords})
	   ? @{$c->{_coords}->saveLines()}
	   : qw()),
	  (defined($c->{_restore})
	   ? @{$c->{_restore}->saveLines()}
	   : qw()),
	 ];
}

##======================================================================
## Subpatch extras
@Pd::Patch::subpatch::ISA = (__PACKAGE__);

##======================================================================
## Pd::Patch::X_restore
##  + subpatch placeholder on parent patch
package Pd::Patch::X_restore;
our @ISA = qw(Pd::Patch::Item);
sub PROPERTIES {
  return [qw(decl type x y name args)];
}
sub DEFAULTS   {
  return {decl=>'#X', type=>'restore', 'x'=>0, 'y'=>0, name=>'', args=>''};
}
__PACKAGE__->init();

##-- aliases
@Pd::Patch::restore::ISA = (__PACKAGE__);

##======================================================================
## Pd::Patch::X_coords
##  + gop properties
package Pd::Patch::X_coords;
our @ISA = qw(Pd::Patch::Item);
sub PROPERTIES {
  return [
	  qw(decl type),
	  qw(xfrom yfrom xto yto xsize ysize), ##-- gop>0
	  #qw(xfrom yfrom x_units_per_pixel y_units_per_pixel xsize ysize), ##-- gop==0
	  qw(gop),             ##-- gop: 0=no, 1=gop[+name] 2=gop[-name]
	  qw(xmargin ymargin), ##-- pd doesn't write these to subpatches...
	 ];
}
##-- defaults:
#0  0 1 1 85 60 0; ##-- gop==0
#0 -1 1 1 85 60 1; ##-- gop==1
sub DEFAULTS   {
  return {
	  decl=>'#X', type=>'coords',
	  xfrom=>0, yfrom=>0, yto=>1, yto=>1, xsize=>85, ysize=>60,
	  gop=>0,                     ##-- gop: 0=no, 1=gop[+name], 2=gop[-name]
	  xmargin=>100, ymargin=>100, ##-- pd doesn't write these to subpatches...
	 };
}
__PACKAGE__->init();

##-- aliases
@Pd::Patch::coords::ISA = (__PACKAGE__);



1; ##-- be happy

__END__

##======================================================================
## Docs
=pod

=head1 NAME

Pd::Patch - Perl parser and wrapper classes for Pd abstractions

=head1 SYNOPSIS

 use Pd::Patch;

 #... stuff happens...

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

not yet written

=cut

##======================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

Pd by Miller Puckette and others.

=head1 AUTHOR

Bryan Jurish E<lt>moocow@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2007, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1),
pd(1).

=cut
